package edu.uci.ics.seal.savasana;

import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;

import edu.uci.ics.seal.savasana.Const.AdaptationApproach;
import edu.uci.ics.seal.savasana.Const.LogLevel;
import edu.uci.ics.seal.savasana.benchmark.SystemUnderTest;
import edu.uci.ics.seal.savasana.runtime.Adaptor;
import edu.uci.ics.seal.savasana.runtime.Registry;

public class Main {

	public static void main(String[] args) {

		Options options = createOptions();
		try {
			CommandLine line = new DefaultParser().parse(options, args);
			processOptions(line);
		} catch (ParseException e) {
			System.out.println(e.getMessage());
			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("savasana", options);
			System.exit(1);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	private static void runExperiments(SystemUnderTest sys, int warmUpTime, int transactionNumbers, double lambda, int speed, AdaptationApproach approach,
			LogLevel logLevel, String outputDir, boolean shutdown) throws InterruptedException {
		Const.OUTPUT_File = outputDir;
		Const.ADAPTATION_APPROACH = approach;
		Const.SUT = sys;
		Const.VERBOSE_LEVEL = logLevel;
		Const.SPEED = speed;
		Const.EARLY_SHUTDOWN = shutdown;
		startRuntimeControl(sys, warmUpTime);
		startSystemsUnderTest(sys, lambda, transactionNumbers);
	}

	private static void startRuntimeControl(SystemUnderTest sys, int warmUpTime) throws InterruptedException {
		Registry.resetRegistry();
		ScheduledExecutorService scheduler = Executors.newScheduledThreadPool(1);
		scheduler.schedule(() -> Adaptor.requestForAdaptation(Util.getAdaptableComponent(sys)), Const.REQUEST_DELAY, TimeUnit.MILLISECONDS);
		scheduler.scheduleWithFixedDelay(() -> Adaptor.checkForComponentsUpdate(), warmUpTime, 10, TimeUnit.MILLISECONDS);
		scheduler.scheduleAtFixedRate(() -> terminateSystem(), warmUpTime * 2, 5000, TimeUnit.MILLISECONDS);
	}

	private static void startSystemsUnderTest(SystemUnderTest sys, double lambda, int transactionNumbers) throws InterruptedException {
		Long[] arrivalTime = Util.getPoissonDistArrivalTime(lambda, transactionNumbers);
		for (int i = 0; i < transactionNumbers; i++) {
			Thread.sleep(arrivalTime[i]);
			Thread myThread = new Thread(sys, Util.generateSystemUnderTestTreadName(sys, i + 1));
			myThread.start();
		}
	}

	private static void terminateSystem() {
		if (Util.checkForTermination()) {
			Util.log(Main.class, "Shutting down");
			try {
				Thread.sleep(Const.WARM_UP_TIME);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			System.exit(0);
		}
	}

	private static Options createOptions() {
		Options options = new Options();
		options.addOption(Option.builder("s").longOpt("system").numberOfArgs(1).required(true).desc("the full package of system under test").build());
		options.addOption(Option.builder("a").longOpt("approach").numberOfArgs(1).required(true).desc("the adaptation approach: Tranquility(t), Savasana (s)")
				.build());
		options.addOption(Option.builder("l").longOpt("log").numberOfArgs(1).desc("log level: (silent(s), quiet(q), concise(c), verbose(v)").build());
		options.addOption(Option.builder("w").longOpt("warmup").type(Number.class).numberOfArgs(1).desc("warm-up time in miliseconds").build());
		options.addOption(Option.builder("t").longOpt("transactions").type(Number.class).numberOfArgs(1)
				.desc("total number of root transactions started in the system under test").build());
		options.addOption(Option.builder("p").longOpt("speed").type(Number.class).numberOfArgs(1).desc("speed of running experiments").build());
		options.addOption(Option.builder("b").longOpt("lamda").type(Number.class).numberOfArgs(1).desc("lamda value used for generating requests").build());
		options.addOption(Option.builder("o").longOpt("output").numberOfArgs(1).desc("output file").build());
		options.addOption(Option.builder("shutdown").desc("shut the systems under tests after adaptation").build());
		return options;
	}

	private static void processOptions(CommandLine line) throws InterruptedException {
		SystemUnderTest sys = Util.findSysUnderTest(line.getOptionValue("s", Const.SUTs[0].getClass().getName()));
		AdaptationApproach approach = AdaptationApproach.getApproach(line.getOptionValue("a").charAt(0));
		LogLevel logLevel = LogLevel.getLoglevel(line.getOptionValue("l").charAt(0));
		int warmUpTime = Integer.valueOf(line.getOptionValue("w", Integer.toString(Const.WARM_UP_TIME)));
		int transactionNumbers = Integer.valueOf(line.getOptionValue("t", Integer.toString(Const.TRANSACTION_NUMBERS)));
		int speed = Integer.valueOf(line.getOptionValue("p", Integer.toString(Const.SPEED)));
		double lambda = Double.valueOf(line.getOptionValue("b", Double.toString(Const.LAMBDA)));
		String outputFile = line.getOptionValue("o");
		boolean shutdown = line.hasOption("shutdown");
		runExperiments(sys, warmUpTime, transactionNumbers, lambda, speed, approach, logLevel, outputFile, shutdown);
	}

}
