package edu.uci.ics.seal.savasana.runtime;

import static edu.uci.ics.seal.savasana.Const.AdaptationApproach.SAVASANA;
import static edu.uci.ics.seal.savasana.Const.AdaptationApproach.TRANQUILITY;
import static edu.uci.ics.seal.savasana.runtime.model.UpdateInterval.Granularity.INTER_COMPONENT;
import static edu.uci.ics.seal.savasana.runtime.model.UpdateInterval.Granularity.INTRA_COMPONENT;
import static edu.uci.ics.seal.savasana.runtime.model.UpdateInterval.IntervalPlace.AFTER;
import static edu.uci.ics.seal.savasana.runtime.model.UpdateInterval.IntervalPlace.BEFORE;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.Signature;
import org.aspectj.lang.annotation.After;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Before;
import org.aspectj.lang.reflect.MethodSignature;

import edu.uci.ics.seal.savasana.Const;
import edu.uci.ics.seal.savasana.Util;
import edu.uci.ics.seal.savasana.benchmark.SystemUnderTest;
import edu.uci.ics.seal.savasana.benchmark.annotations.SafeUpdatePoint;
import edu.uci.ics.seal.savasana.runtime.model.UpdateInterval.Granularity;
import edu.uci.ics.seal.savasana.runtime.model.UpdateInterval.IntervalPlace;

@Aspect
public class Monitor {

	@Before("execution(* edu.uci.ics.seal.savasana.benchmark..*.*(..))")
	public void doBeforeTargetComponentExec(JoinPoint joinPoint) {
		updateRegistryAccordingToSafeAdaptationCriteria(joinPoint.getSignature(), BEFORE);
	}

	@After("execution(* edu.uci.ics.seal.savasana.benchmark..*.*(..))")
	public void doAFterTargetComponentExec(JoinPoint joinPoint) {
		updateRegistryAccordingToSafeAdaptationCriteria(joinPoint.getSignature(), AFTER);
	}

	/**
	 * Check the satisfaction of safe update criteria and update Registry
	 * accordingly for the point of program specified by the given signature.
	 * 
	 * @param signature
	 *            of the current point of the program execution
	 */
	private static void updateRegistryAccordingToSafeAdaptationCriteria(Signature signature, IntervalPlace beforeOrAfter) {
		SystemUnderTest sys = Util.findSysUnderTest(signature);
		if (sys == null)
			return;
		Thread currentThread = Thread.currentThread();
		String componentId = Util.getComponentName(signature, sys);
		String transactionID = signature.getName();
		boolean safeForAdaptation = checkAdaptationCriteria((MethodSignature) signature, beforeOrAfter);
		Registry.updateRegistry(componentId, currentThread, transactionID, beforeOrAfter, safeForAdaptation);
		if (safeForAdaptation) {
			Util.log(Monitor.class, "Safe to adapt " + componentId + " of thread " + currentThread.getName() + " " + beforeOrAfter + " running transactionId "
					+ transactionID);
			// Only Savasana suspends threads (when they are in "safe" state) for adaptation. Another thread checks all threads to see if they are all suspended or not.
			if(Const.ADAPTATION_APPROACH == SAVASANA)
				Adaptor.suspendSafeTransactions(componentId, currentThread);
		}
	}

	/**
	 * Check the satisfaction of adaptation criteria.
	 * 
	 * @param transactionId
	 * @param sys
	 *            the is system under
	 * @return true if the criteria satisfy before/after executing the given
	 *         transactionId of the system under test.
	 */
	private static boolean checkAdaptationCriteria(MethodSignature signature, IntervalPlace beforeOrAfter) {
		Set<Granularity> granularityLevels = new HashSet<Granularity>();
		if (Const.ADAPTATION_APPROACH == TRANQUILITY)
			granularityLevels.add(INTER_COMPONENT);
		if (Const.ADAPTATION_APPROACH == SAVASANA)
			granularityLevels.addAll(Arrays.asList(INTER_COMPONENT, INTRA_COMPONENT));
		return isIntervalSafe(signature, beforeOrAfter, granularityLevels);
	}

	private static boolean isIntervalSafe(MethodSignature signature, IntervalPlace beforeOrAfter, Set<Granularity> granularityLevels) {
		@SuppressWarnings("cast")
		SafeUpdatePoint annotation = (SafeUpdatePoint) signature.getMethod().getAnnotation(SafeUpdatePoint.class);
		if (annotation == null)
			return false;
		return (annotation.beforeOrAfer() == beforeOrAfter && granularityLevels.contains(annotation.adaptaionGranularity()));
	}
}
