package edu.uci.ics.seal.savasana.runtime;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import edu.uci.ics.seal.savasana.Const;
import edu.uci.ics.seal.savasana.Util;
import edu.uci.ics.seal.savasana.Const.AdaptationApproach;
import edu.uci.ics.seal.savasana.runtime.model.AdaptationLog;
import edu.uci.ics.seal.savasana.runtime.model.RegistryKey;
import edu.uci.ics.seal.savasana.runtime.model.RegistryValue;
import edu.uci.ics.seal.savasana.runtime.model.UpdateInterval.IntervalPlace;

public class Registry {

	static Map<RegistryKey, RegistryValue> registry = new HashMap<RegistryKey, RegistryValue>();
	static Map<String, AdaptationLog> reachabilityLogs = new HashMap<String, AdaptationLog>();
	static Map<String, AdaptationLog> disruptionLogs = new HashMap<String, AdaptationLog>();
	static List<Long> reachabilityTimes = new ArrayList<>();
	static List<Long> disruptionTimes = new ArrayList<>();

	public static void resetRegistry() {
		Util.log(Registry.class, "Resetting Registry...");
		registry = new HashMap<RegistryKey, RegistryValue>();
		reachabilityLogs = new HashMap<String, AdaptationLog>();
		disruptionLogs = new HashMap<String, AdaptationLog>();
	}

	static synchronized void updateRegistry(String componentId, Thread thread, String transactionId, IntervalPlace beforeOrAfter, boolean isSafe) {
		RegistryKey regKey = new RegistryKey(componentId, thread);
		RegistryValue regVal = new RegistryValue(isSafe, transactionId, componentId, beforeOrAfter);
		registry.put(regKey, regVal);
	}

	static synchronized void logThreadSuspension(Thread thread) {
		AdaptationLog adaptationLog = disruptionLogs.get(thread.getName());
		if (adaptationLog == null)
			adaptationLog = new AdaptationLog();
		adaptationLog.logRequest();
		disruptionLogs.put(thread.getName(), adaptationLog);
	}

	static synchronized void logRequestForAdaptation(String componentId) {
		AdaptationLog adaptationLog = reachabilityLogs.get(componentId);
		if (adaptationLog == null)
			adaptationLog = new AdaptationLog();
		adaptationLog.logRequest();
		reachabilityLogs.put(componentId, adaptationLog);
	}

	static void logSuccessfulAdaptation(String componentId, List<Thread> threadSet) {
		AdaptationLog adaptationLog = reachabilityLogs.get(componentId);
		if (adaptationLog != null)
			adaptationLog.logAdaptation();
		for (Thread thread : threadSet) {
			adaptationLog = disruptionLogs.get(thread.getName());
			if (adaptationLog != null)
				adaptationLog.logAdaptation();
		}
	}

	static boolean isSafeToUpdate(String componentID, Thread thread) {
		RegistryValue regVal = registry.get(new RegistryKey(componentID, thread));
		if (regVal == null)
			return true;
		return regVal.isSafe();
	}

	public static Long calcReachabilityTime() {
		String componentId = reachabilityLogs.keySet().toArray(new String[1])[0];
		if (componentId == null)
			return null;
		Long reachability = reachabilityLogs.get(componentId).getWaitTime().stream().filter(Objects::nonNull).mapToLong(Long::longValue).sum();
		reachabilityTimes.add(reachability);
		return reachability;
	}

	public static Long calcDisruptionTime() {
		Long disruption = disruptionLogs.values().stream().map(AdaptationLog::getWaitTime).filter(Objects::nonNull).flatMap(List::stream)
				.mapToLong(Long::longValue).sum();
		disruptionTimes.add(disruption);
		return disruption;
	}

	public static String registeryToString(String componentID) {
		StringBuilder sb = new StringBuilder();
		for (RegistryKey key : registry.keySet().stream().filter(key -> key.getComponentId().equals(componentID))
				.sorted((key1, key2) -> key1.getThreadName().compareTo(key2.getThreadName())).collect(Collectors.toList()))
			sb.append(key + "->" + registry.get(key) + ",");
		return "Safe Intervals: " + sb.toString();
	}

	public static String getFormattedExprimentsResult(String system, AdaptationApproach approach, Number reachability, Number disruption, boolean fullDesc) {
		StringBuilder sb = new StringBuilder();
		if (fullDesc) {
			sb.append("\n------------------------------------------- Experiment Results -----------------------------------------------------\n");
			sb.append(String.format("%-20s%-20s%-20s%s\n", "System", "Approach", "Reachability(ms)", "Disruption(ms)"));
			sb.append(String.format("%-20s%-20s%-20d%d\n", system, approach, reachability, disruption));
			sb.append("---------------------------------------------------------------------------------------------------------------------\n");
		} else {
			sb.append(String.format("%s,%s,%d,%d\n", system, approach, reachability, disruption));
		}
		return sb.toString();
	}

	public static void printExprimentsResult() {
		String outputFile = Const.OUTPUT_File;
		boolean fileOutput = outputFile != null;
		String result = getFormattedExprimentsResult(Util.getSystemUnderTestShortName(Const.SUT), Const.ADAPTATION_APPROACH, calcReachabilityTime(),
				calcDisruptionTime(), !fileOutput);
		Util.log(Adaptor.class, result, Const.LogLevel.SILENT, false);
		if (fileOutput) {
			File out = new File(outputFile);
			try (Writer writer = new BufferedWriter(new FileWriter(out, true))) {
				if(!out.exists() || out.length()==0 )
					writer.write("System, Approach, Reachability(ms), Disruption(ms)\n");
				writer.write(result);
				writer.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

}
